import os


def load_commands(directory):
    for filename in os.listdir(directory):
        if filename.endswith(".py"):
            exec(f"from {directory.replace('/', '.')}.{filename[:-3]} import *")
        elif os.path.isdir(f"{directory}/{filename}") and not filename.endswith("__"):
            load_commands(f"{directory}/{filename}")



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelType(id, context):
    """
    Ex. $channelType[ChannelID]
    returns the channel type of the given args
    """
    if len(id) < 1:
        CharmCordErrors("No parameter provided for '$channelType'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(id.replace("<#", "").replace(">", ""))
        return channel.type
    except ValueError:
        CharmCordErrors(f"$channelType: {id} not valid channel id\nCommand: {context.command.name}")



from CharmCord.tools import checkArgCheck, checkArgs, findBracketPairs, noArguments, lets, isValid
from CharmCord.globeHandler import get_globals

########################################
#              COMMANDS                #
########################################


class Commands:
    # Global variables

    @staticmethod
    def command(name: str, code: str, aliases: list = [], bot=None):
        # Define command function dynamically

        @bot.command(name=name, aliases=aliases)
        async def go(ctx, *args, codes=code):
            funcs = get_globals()[0]
            new_code = await checkArgCheck(args, codes, ctx)
            if new_code == "Failed":
                return
            code1 = await noArguments(new_code, funcs, ctx)
            code2 = checkArgs(args, code1)
            final_code = await isValid(code2, funcs)
            await findBracketPairs(final_code, funcs, ctx)
            if len(lets) >= 1:
                lets.clear()



async def hasPerm(args, context):
    vals = args.split(";")
    user = vals[0].replace("<@", "").replace(">", "")
    perm = vals[1].lower()
    guild = await context.guild.fetch_member(int(user))
    perms = [i for i in guild.guild_permissions]
    for i in perms:
        if perm in i:
            return True
    else:
        return False



import discord
import CharmCord.CharmErrorHandling as ErrorHandling
from CharmCord.globeHandler import get_globals

EH = ErrorHandling.CharmErrorHandling()


async def guildTextChannels(id, context):
    bots = get_globals()[1]

    try:
        id = int(id)
        text = []
        guild = await bots.fetch_guild(id)
        channels = list(await guild.fetch_channels())
        for channel in channels:
            if isinstance(channel, discord.TextChannel):
                text.append(channel.name)
        return text
    except ValueError:
        EH.Errors(2, "None")



import random

import discord
from discord.ext import tasks

from CharmCord.CharmErrorHandling import CharmCordErrors


def setActivity(message: str, typing: str = "watching"):
    if typing.lower() not in ["watching", "playing", "listening"]:
        CharmCordErrors("Unknown Status type used for bot activity")
    if typing.lower() == "watching":
        act = discord.Activity(type=discord.ActivityType.watching, name=message)
    elif typing.lower() == "playing":
        act = discord.Activity(type=discord.ActivityType.playing, name=message)
    else:
        act = discord.Activity(type=discord.ActivityType.listening, name=message)

    return act


def loopActivity(time: int, messages: [str], types: str = "watching"):
    @tasks.loop(seconds=time)
    async def updateActivity(message=messages, typing=types):
        from CharmCord.utils.CharmCord import bots
        await bots.change_presence(
            activity=discord.Activity(
                type=discord.ActivityType.watching, name=random.choice(message)
            )
        )

    loopAct = updateActivity
    return loopAct
# Still having a lot of errors with this, will return when I have the time to set it up



import asyncio

from CharmCord.CharmErrorHandling import CharmCordErrors


async def waitReaction(args, context):
    """
    Ex. $waitMessage[ChannelID;User;emoji;timeout;timeoutErrMsg]
    """
    from CharmCord.utils.CharmCord import bots
    split = args.split(";")
    if len(split) < 3:
        raise SyntaxError("args, User, or timeout not provided to $waitMessage")
    try:
        channel_id = split[0]
        users = split[1]
        emoji = split[2]
        timeout = int(split[3])

        def check(rct, usr):
            if int(channel_id) == context.channel.id:
                if users == "everyone":
                    if str(rct.emoji.name) == emoji:
                        return True
                elif int(users) == usr.id:
                    if str(rct.emoji.name) == emoji:
                        return True

        error = None
        if len(split[4]) > 1:
            error = split[4]
        if error is None:
            reaction, user = await bots.wait_for("reaction_add", timeout=timeout, check=check)
            try:
                return reaction.emoji.name
            except AttributeError:
                return reaction.emoji
        else:
            try:
                reaction, user = await bots.wait_for("reaction_add", timeout=timeout, check=check)
                try:
                    return reaction.emoji.name
                except AttributeError:
                    return reaction.emoji
            except asyncio.TimeoutError:
                await context.channel.send(error)
    except ValueError:
        CharmCordErrors(f"args Error in {context.command.name}")
        return



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelNsfw(id, context):
    """
    Ex. $channelNsfw[ChannelID]
    returns a boolean about a channel nsfw setting
    """
    if len(id) < 1:
        CharmCordErrors("No parameter provided for '$channelNsfw'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(id.replace("<#", "").replace(">", ""))
        if channel.is_nsfw():
            return True
        else:
            return False
    except ValueError:
        CharmCordErrors(f"id ({id}) is an invalid Channel ID")



from CharmCord.tools import checkArgCheck, checkArgs, findBracketPairs, noArguments, isValid
from CharmCord.globeHandler import get_globals

AC = {}


class CharmCogs:

    @staticmethod
    def command_cogs(name, code):
        funcs = get_globals()[0]
        bots = get_globals()[1]

        @bots.command(name=name)
        async def go(ctx, *args, codes=code):
            context = ctx
            new_code = await checkArgCheck(args, codes, context)
            if new_code == "Failed":
                return
            code1 = await noArguments(new_code, funcs, context)
            code2 = checkArgs(args, code1)
            final_code = await isValid(code2, funcs)
            await findBracketPairs(final_code, funcs, context)

    @staticmethod
    def slashcommand_cogs(name, code, args: list[dict], description):

        def slash_command():
            bots = get_globals()[1]
            types = {1: "str", 2: "int"}
            new_args = []
            arg_descripts = []
            for i in args:
                if len(i) != 0:
                    try:
                        i['description']
                    except KeyError:
                        i['description'] = "No Description"
                    new_args.append(f"{i['name']}: {types[i['type']]}")
                    arg_descripts.append(f"{i['name']} ({types[i['type']]}): {i['description']}")
            nl = "\n\t\t\t\t\t\t\t\t"
            needs = {"arguments": args, "codes": code, "bot": bots, "name": name}
            func = f"""@bot.tree.command(name=name)
async def go(ctx, {', '.join(new_args)}):
                            '''{description}

                            Args:
                                {nl.join(arg_descripts)}
                            '''
                            from CharmCord.globeHandler import get_globals
                            from CharmCord.tools import noArguments, slashArgs, findBracketPairs, lets
                            funcs = get_globals()[0]
                            context = ctx
                            new = []
                            for i in arguments:
                                new.append(eval(i['name']))
                            finalCode = await noArguments(codes, funcs, context)
                            finalCode = slashArgs(new, finalCode)
                            await findBracketPairs(finalCode, funcs, context)
                            if len(lets) >= 1:
                                lets.clear()
                    """
            exec(func, needs)

        slash_command()



async def If(args, context):
    choices = ["==", ">=", "<=", "!=", "<", ">"]
    operator_mapping = {
        "==": lambda x, y: x == y,
        "!=": lambda x, y: x != y,
        ">": lambda x, y: x > y,
        ">=": lambda x, y: x >= y,
        "<": lambda x, y: x < y,
        "<=": lambda x, y: x <= y,
    }
    for i in choices:
        if i in args:
            if i in ["==", "!="]:
                vals = args.split(i)
                val1 = vals[0].strip()
                val2 = vals[1].strip()
            else:
                vals = args.split(i)
                val1 = int(vals[0])
                val2 = int(vals[1])
            result = operator_mapping.get(i, lambda x, y: None)(val1, val2)
            if result:
                return True
            else:
                return False
    return False



"""
CharmCord | Discord.py Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper of discord.py to make developing discord bots easier for new developers

:copyright: (c) 2022-Present LilbabJJx-1 & (c) 2023-present CharmCord Development
:license: MIT, see LICENSE for more details.
"""

from .utils import *
from .functions import *



async def editMessage(args, context):
    if ";" in args:
        values = args.split(";")
        try:
            msg_id, new_msg = values[0], values[1]
            msg = await context.channel.fetch_message(int(msg_id))
            if msg is None:
                raise SyntaxError(
                    f"$editMessage Message not found with ID: {msg_id}\nCommand: '{context.command.name}'")
            await msg.edit(content=new_msg)
            return
        except IndexError:
            raise SyntaxError(f"$editMessage requires 2 parameters in command: '{context.command.name}'")
        except ValueError:
            raise SyntaxError(f"$editMessage Invalid ID provided in command: '{context.command.name}'")



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelChangedRoles(args, context):
    """
    Ex. $channelChangedRoles[ChannelID;Index]
    """
    if len(args) < 1:
        CharmCordErrors("No parameter provided for '$channelChangedRoles'")
        return
    bots = get_globals()[1]

    if ";" in args:
        args = args.split(";")
        args = args[0]
        index = args[1]
        try:
            int(args)
            channel = await bots.fetch_channel(args)
            return channel.changed_roles[int(index) - 1]
        except ValueError:
            CharmCordErrors("No parameter provided for '$channelCategoryRoles'")
    else:
        try:
            int(args)
            channel = await bots.fetch_channel(args)
            return channel.changed_roles
        except ValueError:
            CharmCordErrors("No parameter provided for '$channelCategoryRoles'")



from CharmCord.all_functions import newline_char
from ._btnOpts_ import views
from CharmCord.globeHandler import get_globals


async def sendMessage(args: str, context):
    bots = get_globals()[1]
    split = args.split(";")
    if len(split) < 2:
        raise SyntaxError("args or message not provided to $sendMessage")
    try:
        channel_id = split[0]
        channel = await bots.fetch_channel(int(channel_id))
        if len(split) > 1:
            message = split[1]
            message = message.replace(newline_char, "\n")
            if len(views) > 0:
                sent = await channel.send(message, view=views[0])
                return sent.id
            else:
                sent = await channel.send(message)
                return sent.id
        if len(views) > 0:
            sent = await channel.send(view=views)
            return sent.id
    except Exception as e:
        print(e)
        raise SyntaxError("Can't send empty message!")



# TODO: Add more functions to be used
from .Channels import *
from .Client import *
from .Events import *
from .Guilds import *
from .Messages import *
from .Python import *
from .Users import *
from .Variables import *
from .clientActivity import setActivity


class SlashCommands:

    @staticmethod
    def slash_command(name, code, args: list[dict] = None, description=None, bot=None):
        types = {1: "str", 2: "int"}
        new_args = []
        arg_descripts = []
        if args is not None:
            for i in args:
                if len(i) != 0:
                    try:
                        i['description']
                    except KeyError:
                        i['description'] = "No Description"
                    new_args.append(f"{i['name']}: {types[i['type']]}")
                    arg_descripts.append(f"{i['name']} ({types[i['type']]}): {i['description']}")
                    needs = {"arguments": args, "codes": code, "bot": bot, "name": name}
        else:
            needs = {"codes": code, "bot": bot, "name": name}
        nl = "\n\t\t\t"
        func = f"""
@bot.tree.command(name=name)
async def go(ctx, {', '.join(new_args)}):
                '''{description}
                
                Args:
                    {nl.join(arg_descripts)}
                '''
                from CharmCord.globeHandler import get_globals
                from CharmCord.tools import noArguments, slashArgs, findBracketPairs, lets
                funcs = get_globals()[0]
                context = ctx
                new = []
                for i in arguments:
                    new.append(eval(i['name']))
                finalCode = await noArguments(codes, funcs, context)
                finalCode = slashArgs(new, finalCode)
                await findBracketPairs(finalCode, funcs, context)
                if len(lets) >= 1:
                    lets.clear()
        """
        func2 = f"""
@bot.tree.command(name=name)
async def go(ctx, {', '.join(new_args)}):
                '''{description}
                '''
                from CharmCord.globeHandler import get_globals
                from CharmCord.tools import noArguments, slashArgs, findBracketPairs, lets
                funcs = get_globals()[0]
                context = ctx
                new = []
                finalCode = await noArguments(codes, funcs, context)
                await findBracketPairs(finalCode, funcs, context)
                if len(lets) >= 1:
                    lets.clear()
        """
        if args is None:
            exec(func2, needs)
        else:
            exec(func, needs)



from discord.ui import View, Button
from ._btnOpts_ import button, views
from CharmCord.globeHandler import get_globals


async def addButton(args, ctx):
    from CharmCord.tools import checkArgCheck, checkArgs, findBracketPairs, noArguments, lets, isValid
    name, custom_id = args.split(";")

    class MyView(View):
        def __init__(self):
            super().__init__()
            self.all_buttons = {}

        def add_button(self, button_name, button_instance):
            self.all_buttons[button_name] = button_instance
            return

        def get_button(self, button_name):
            if button_name in self.all_buttons:
                return self.all_buttons[button_name]
            return None

    if len(views) == 0:
        async def button_go(button_interaction):
            funcs = get_globals()[0]
            views.clear()
            try:
                codes = button[custom_id]
            except KeyError:
                raise Exception(f"There's no code for button: {custom_id}")
            new_code = await checkArgCheck(args, codes, button_interaction)
            if new_code == "Failed":
                return
            code1 = await noArguments(new_code, funcs, button_interaction)
            code2 = checkArgs(args, code1)
            final_code = await isValid(code2, funcs)
            await findBracketPairs(final_code, funcs, button_interaction)
            if len(lets) >= 1:
                lets.clear()

        data = Button(label=name, custom_id=custom_id)
        mine = MyView()
        mine.add_button(name, data)
        views.append(mine)
        data.callback = button_go
        views[0].add_item(data)
    else:
        async def button_go(button_interaction):
            funcs = get_globals()[0]
            views.clear()
            codes = button[custom_id]
            new_code = await checkArgCheck(args, codes, button_interaction)
            if new_code == "Failed":
                return
            code1 = await noArguments(new_code, funcs, button_interaction)
            code2 = checkArgs(args, code1)
            final_code = await isValid(code2, funcs)
            await findBracketPairs(final_code, funcs, button_interaction)
            if len(lets) >= 1:
                lets.clear()

        data = Button(label=name, custom_id=custom_id)
        data.callback = button_go
        views[0].add_item(data)
    return



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelCreated(args: str, context, timezones, format_datetime):
    """
    Ex. $channelCreated[ChannelID]
    """
    if len(args) < 1:
        raise CharmCordErrors("No parameter provided for '$channelCreated'")
    est, utc, pst = timezones
    try:
        ids, time, form = tuple(args.split(";"))
        time = locals()[time.strip()]
        form = form.lower()
    except ValueError:
        form = "full"
        try:
            ids, time = tuple(args.split(","))
            time = locals()[time.strip()]
        except:
            ids = args
            time = utc

    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(ids.replace("<#", "").replace(">", ""))

        desiredDateForm = format_datetime(channel.created_at, form, time)
        if desiredDateForm != "ERROR":
            return desiredDateForm
        else:
            raise SyntaxError("Invalid Format option in $channelCreated!")

    except ValueError:
        CharmCordErrors(f"$channelCreated: {ids} not valid channel id\nCommand: {context.command.name}")



async def count(code, context):
    if isinstance(code, list):
        try:
            return len(code)
        except:
            return
    else:
        try:
            return len([i for i in code.split(" ") if len(i) > 0])
        except:
            return 0



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelPosition(id, context):
    """
    Ex. $channelPosition[ChannelID]
    returns the position of a given channel args
    """
    if len(id) < 1:
        CharmCordErrors("No parameter provided for '$channelPosition'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(id.replace("<#", "").replace(">", ""))
        return channel.position
    except ValueError:
        CharmCordErrors(f"$channelPosition: {id} not valid channel id\nCommand: {context.command.name}")



async def divide(args, context):
    if ";" in args:
        values = args.split(";")
        try:
            val1 = float(values[0])
            val2 = float(values[1])
            new = str(val1 / val2)
            if new.endswith(".0"):
                new = new.replace(".0", "")
            return new
        except ValueError:
            raise SyntaxError("$divide parameters can only can numbers")
        except IndexError:
            raise SyntaxError("$divide requires 2 parameters")
    else:
        raise SyntaxError("$divide requires 2 parameters")


import discord
from CharmCord.all_functions import newline_char
from CharmCord.functions.Messages._btnOpts_ import views


async def buttonSend(args, interaction: discord.Interaction):
    try:
        message = args
        message = message.replace(newline_char, "\n")
        if len(views) > 0:
            await interaction.response.send_message(message, view=views[0])
            return
        await interaction.response.send_message(message)
    except Exception as e:
        raise SyntaxError("Can't send empty message!")
    return



import discord
from CharmCord.globeHandler import get_globals
from CharmCord.all_functions import newline_char


async def sendEmbed(args: str, context):
    """
    Ex. $sendEmbed[Channel args;Title;Message;Color?;image?;Footer?]
    Send an Embed
    """
    bots = get_globals()[1]
    args = args.replace(newline_char, "\n")
    split = args.split(";")
    try:
        channel_id = split[0]
        title = split[1]
        message = split[2]
        color = split[3]
        image = split[4]
        footer = split[5]
        channel = await bots.fetch_channel(int(channel_id))
        embed = eval(
            f"discord.Embed(title=title, description=message, color=discord.Color.{color.lower()}())"
        )
        embed.set_footer(text=footer)
        embed.set_image(url=image)

        if isinstance(context, discord.Interaction):
            try:
                await context.response.send_message(embed=embed)
                message = await context.original_response()
                return message.id
            except:
                await context.followup.send(embed=embed)
                message = await context.original_response()
                return message.id
        message = await channel.send(embed=embed)
    except discord.ClientException:
        raise SyntaxError("Can't send empty message!")
    except IndexError:
        raise SyntaxError("Not enough arguments in $sendEmbed")
    except SyntaxError:
        channel_id = split[0]
        title = split[1]
        message = split[2]
        image = split[4]
        footer = split[5]
        channel = await bots.fetch_channel(int(channel_id))
        embed = eval("discord.Embed(title=title, description=message)")
        embed.set_footer(text=footer)
        embed.set_image(url=image)
        if isinstance(context, discord.Interaction):
            try:
                await context.response.send_message(embed=embed)
                message = await context.original_response()
                return message.id
            except:
                await context.followup.send(embed=embed)
                message = await context.original_response()
                return message.id
        message = await channel.send(embed=embed)
    return message.id



import json


async def setServerVar(args, context):
    ag = args.split(";")
    server = ag[0]
    var = ag[1]
    value = ag[2]
    with open("variables.json", "r") as variables:
        total = json.load(variables)
    with open("variables.json", "w") as variables:
        try:
            total.update({f"{server}_{var}": int(value)})
        except ValueError:
            total.update({f"{server}_{var}": value})
        json.dump(total, variables)
    return



import CharmCord.CharmErrorHandling as ErrorHandling

EH = ErrorHandling.CharmErrorHandling()


async def guildID(empty, context):
    """Get guild id of the guild where command was invoked"""
    return context.guild.id



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def sendDM(args, context):
    bots = get_globals()[1]
    if ";" not in args:
        CharmCordErrors(f"Not enough arguments in $sendDM: {context.command.name} command")

    values = args.split(";")
    user = values[0].replace("<@", "").replace(">", "")
    message = values[1]
    dm = await bots.fetch_user(user)
    await dm.send(message)
    return



import aiohttp

from CharmCord.CharmErrorHandling import CharmCordErrors


async def getJson(args, context):
    """
    Ex. $getJson[url;keyValue]
    :param args:
    :param context:
    :return:
    """
    if ";" not in args:
        CharmCordErrors(f"Not enough args in $getJson: {context.command.name} command")
    vals = args.split(";")
    url = vals[0]
    key = vals[1]

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status == 200:
                data = await response.json()
                try:
                    return data[key]
                except KeyError:
                    CharmCordErrors(f"Unknown key in $getJson: {context.command.name} command")



import os

current_directory = os.path.dirname(__file__)
files = os.listdir(current_directory)
python_files = [file for file in files if file.endswith('.py') and "_" not in file]
for file in python_files:
    try:
        module_name = file[:-3]  # Remove the '.py' extension
        exec(f'from .{module_name} import {module_name}')  # nosec
    except ModuleNotFoundError:
        continue




async def deleteMessage(args, context):
    try:
        msg_id = args
        msg = await context.channel.fetch_message(int(msg_id))
        if msg is None:
            raise SyntaxError(f"$deleteMessage Message not found with ID: {msg_id}\nCommand: '{context.command.name}'")
        await msg.delete()
        return
    except ValueError:
        raise SyntaxError(f"$deleteMessage Invalid ID provided in command: '{context.command.name}'")



from CharmCord.globeHandler import get_globals


async def message(emp, context):
    bots = get_globals()[1]
    remove = ""
    if str(bots.command_prefix) in context.message.content:
        for i in context.message.content:
            if i == " ":
                break
            else:
                remove += i
    return str(context.message.content).replace(remove, "")



import asyncio
from CharmCord.globeHandler import get_globals
from CharmCord.CharmErrorHandling import CharmCordErrors


async def waitMessage(args, context):
    """
    Ex. $waitMessage[ChannelID;User;timeout;timeoutErrMsg]
    """
    bots = get_globals()[1]
    split = args.split(";")
    if len(split) < 3:
        raise SyntaxError("args, User, or timeout not provided to $waitMessage")
    try:
        channel_id = split[0]
        user = split[1]
        timeout = int(split[2])

        def check(msg):
            if int(channel_id) == msg.channel.id:
                if user == "everyone":
                    return True
                elif int(user) == msg.author.id:
                    return True

        error = None
        if len(split[3]) > 1:
            error = split[3]
        if error is None:
            try:
                message = await bots.wait_for("message", timeout=timeout, check=check)
                return message.content
            except asyncio.TimeoutError:
                return
        else:
            try:
                message = await bots.wait_for("message", timeout=timeout, check=check)
                return message.content
            except asyncio.TimeoutError:
                await context.channel.send(error)
    except ValueError:
        CharmCordErrors(f"args Error in {context.command.name}")
        return



from CharmCord.CharmErrorHandling import CharmCordErrors


async def mentions(ids, context):
    try:
        return context.message.mentions[int(ids) - 1].id
    except ValueError:
        CharmCordErrors("Invalid args in $mentions | Command..")
    except IndexError:
        return None




TotalFuncs = None
bots = None
all_vars = None
lets = {}


def update_globals(key: str, value):
    global TotalFuncs
    global bots
    global all_vars
    global lets
    if key.lower().startswith("t"):
        TotalFuncs = value
    elif key.lower().startswith("b"):
        bots = value
    elif key.lower().startswith("l"):
        lets = value
    else:
        all_vars = value


def get_globals():
    return [TotalFuncs, bots, all_vars, lets]



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelCategoryID(args: str, context):
    """
    Ex. $channelCategoryID[ChannelID]
    Returns the args of the current category args
    """
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(args.replace("<#", "").replace(">", ""))
        return channel.category.id
    except ValueError:
        CharmCordErrors(f"$channelCategoryID: {args} not valid channel id\nCommand: {context.command.name}")



from datetime import datetime as d_t
from pytz import timezone
from CharmCord.all_functions import date_funcs, ifse, all_Funcs, no_arg_Funcs
from CharmCord.functions import *

timezones = (timezone("EST"), timezone("UTC"), timezone("US/Pacific"))
lets = {}


class FunctionHandler:
    def __init__(self):
        self.funcs = {}

    def register_functions(self):
        for line in all_Funcs:
            function = eval(line.replace("$", ""))  # nosec
            self.funcs[line.replace("\n", "").lower()] = function
            continue

    async def execute_functions(self, keyword, args, context):
        if keyword in ifse:
            check = await self.funcs[keyword](args, context)
            return check
        if keyword in date_funcs:
            return await self.funcs[keyword](args, context, timezones, format_datetime)
        return await self.funcs[keyword](args, context)


async def noArguments(entry: str, functions, context):
    for func in no_arg_Funcs:
        if func in entry:
            entry = entry.replace(
                func,
                str(await functions.execute_functions(func.lower(), None, context)),
            )

    return entry


def slashArgs(args, code):
    if "$slashArgs" in code:
        while "$slashArgs" in str(code):
            count = 0
            end = None
            balance = 0
            start = code.index("$slashArgs") + 10
            look = code[start: len(code)]
            for char in look:
                if char == "[":
                    start = count
                    count += 1
                    balance += 1
                    continue
                if char == "]":
                    end = count
                    balance -= 1
                count += 1
                if balance == 0 and start is not None and end is not None:
                    try:
                        # Replace $args with arguments
                        code = str(code).replace(
                            f"$slashArgs[{look[start + 1:end]}]",
                            str(args[int(look[start + 1: end]) - 1]),
                        )
                        break
                    except IndexError:
                        raise SyntaxError(
                            f"$slashArgs[{int(look[start + 1:end])}] Not Provided"
                        )
    return code


EndIf = True


async def findBracketPairs(entry: str, functions, context):
    """
    `findBracketPairs` is used to look through the text and determine what function
    is holding what information and how to reformat it into a way for the Discord.py
    package to read it.
    :param entry: The string text of the command
    :param functions: List of all possible functions to use
    :param context: Discord context
    :return: Awaited Async functions
    """
    global EndIf
    test = [line.strip() for line in entry.split("\n") if len(line.strip()) > 0]
    starts, pairs, index = 0, 0, 0
    new = []
    for char in test:
        pairs += char.count("[")
        pairs -= char.count("]")
        if starts >= len(test):
            starts -= 1
        if char.strip().startswith("$") and pairs > 0 and len(new) == 0:
            new.append(test[test.index(char)] + " " + test[test.index(char) + 1])
            starts += 1
            continue
        if starts == 1:
            starts += 1
            continue
        if char.strip().startswith("$") and pairs == 0 and len(new) == 0:
            new.append(test[test.index(char)])
            index += 1
        elif char.strip().startswith("$") and pairs == 0 and len(new) > 0:
            new.append(test[test.index(char)])
            index += 1
        elif char.strip().startswith("$") and pairs > 0 and len(new) > 0:
            new.append(test[test.index(char)])
            index += 1
        elif char.strip()[0] != "$" and pairs == 0 and len(new) > 0:
            new[index] = new[index] + " " + test[test.index(char)]
        elif char.strip()[0] != "$" and pairs > 0 and len(new) > 0:
            new[index] = new[index] + " " + test[test.index(char)]

    if len(new) == 0:
        test = [line.strip() for line in entry.split("\n") if len(line.strip()) >= 3]
    line = 0
    for code in new:
        line += 1
        if EndIf:
            if code.strip().startswith("$end"):
                return
            if code.strip().lower().startswith("$onlyif") and line != 1:
                raise SyntaxError("$OnlyIf should only be at the beginning of a command")
            if code.strip().lower().startswith("$endif"):
                continue
            elif code.strip().lower().startswith("$elif"):
                if not any("$if" in Char.lower() for Char in test):
                    raise SyntaxError("No $If found in command before $ElIf")
                EndIf = False
                continue
            else:
                pass
        else:
            if code.strip().lower().startswith("$elif"):
                EndIf = True
            elif code.strip().lower().startswith("$endif"):
                EndIf = True
                continue
            else:
                continue
        code = code.strip()
        first = None
        last = None
        count = 0
        balance1 = 0
        for char in code:
            if char == "[" and first is None:
                first = count
                balance1 += 1
                count += 1
                continue
            if char == "[":
                balance1 += 1
            elif char == "]":
                last = count
                balance1 -= 1
            if first is not None and last is not None and balance1 == 0:
                break
            count += 1
        argument = str(code[first + 1: last])
        keyword = code[0:first]
        find = [first, last, keyword, argument, context]
        func_job = False
        while "[" in str(argument) and "]" in str(argument) and "$" in str(argument) and func_job is False:
            if "$charmai" in argument.lower():
                func_job = True
            count = 0
            start = None
            end = None
            balance = 0
            for char in argument:
                digits = ["1", "2", "3", "4", "5", "6", '7', '8', "9", "0"]  # A keyword will never start or have a
                # digit in it
                if char == "$" and start is None and argument[count + 1] != "$" and argument[count + 1] not in digits:
                    # $$keyword will discount the first $ as part of the text
                    start = count
                elif char == "[":
                    balance += 1
                elif char == "]":
                    end = count
                    balance -= 1
                count += 1
                if balance == 0 and start is not None and end is not None:
                    break
            if start != 0:
                argument = (
                        argument[:start]
                        + str(await findBracketPairs(argument[start: end + 1], functions, context))
                        + argument[end + 1:]
                )
            else:
                argument = (
                        str(await findBracketPairs(argument, functions, context))
                        + argument[end + 1:]
                )
            find = [first, last, keyword, argument, context]
        if find[2].lower() in functions.funcs:
            response = await functions.execute_functions(find[2].lower(), find[3], find[4])
            if find[2].lower() == "$onlyif" and response is False:
                return
            if find[2].lower() == "$if" and response is False:
                EndIf = False
                if not any("$endif" in Char.lower() for Char in test):
                    raise SyntaxError("No $EndIf found in command after $If")
            elif find[2].lower() == "$if":
                if not any("$endif" in Char.lower() for Char in test):
                    raise SyntaxError("No $EndIf found in command after $If")

                continue
            if find[2].lower() == "$elif" and EndIf is False:
                if not any("$if" in Char.lower() for Char in test):
                    raise SyntaxError("No $If found in command before $ElIf")

                if not any("$endif" in Char.lower() for Char in test):
                    raise SyntaxError("No $EndIf found in command after $Elif")

            if find[2].lower() == "$elif":
                if not any("$if" in Char.lower() for Char in test):
                    raise SyntaxError("No $If found in command before $ElIf")

                if not any("$endif" in Char.lower() for Char in test):
                    raise SyntaxError("No $EndIf found in command after $Elif")

            EndIf = response is not False

        else:
            response = find[2]

    try:
        return response
    except Exception as e:
        raise Exception(f"Error at: {e}")


def checkArgs(args, code):
    if "$args" in code:
        while "$args" in str(code):
            if "$args[" in code:
                count = 0
                end = None
                balance = 0
                start = code.index("$args") + 5
                look = code[start: len(code)]
                for Char in look:
                    if Char == "[":
                        start = count
                        count += 1
                        balance += 1
                        continue
                    if Char == "]":
                        end = count
                        balance -= 1
                    count += 1
                    if balance == 0 and start is not None and end is not None:
                        try:
                            # Replace $args with arguments
                            code = str(code).replace(
                                f"$args[{look[start + 1:end]}]",
                                args[int(look[start + 1: end]) - 1],
                            )
                            break
                        except IndexError:
                            raise SyntaxError(
                                f"$args[{int(look[start + 1:end])}] Not Provided"
                            )
            else:
                add = [char for char in args]
                code = str(code).replace(f"$args", str(add))
    return code


async def isValid(code, functions):
    if "$isValidFunc" in code:
        while "$isValidFunc" in code:
            start = code.index("$isValidFunc[") + 13
            area = code[start:]
            if "$" not in area[:area.index(']')]:
                valid = str(f"${area[:area.index(']')]}").lower() in functions.funcs
                code = code.replace(
                    f"$isValidFunc[{area[:area.index(']')]}]", str(valid)
                )
            else:
                valid = str(f"{area[:area.index(']')]}").lower() in functions.funcs
                code = code.replace(
                    f"$isValidFunc[{area[:area.index(']')]}]", str(valid)
                )
            return code
    return code


async def checkArgCheck(args, code, context):
    if "$argcheck" in code.lower():
        if code.lower().count("$argcheck") > 1:
            raise Exception("Too many $argCheck in a single command | Max is 1!")
        start = code.lower().index("$argcheck[") + 10
        area = code[start:]
        try:
            if ";" in area[: area.index("]")]:
                arg_total = area[: area.index(";")]
                warning = area[area.index(";") + 1: area.index("]")]
                if len(args) < int(arg_total):
                    await context.channel.send(warning)
                    return "Failed"
                code = code.replace(
                    f"$argCheck[{arg_total}{area[area.index(';'):area.index(']')]}]\n",
                    "",
                )
                return code
            else:
                arg_total = area[: area.index("]")]
                if len(args) < int(arg_total):
                    return "Failed"
                code = code.replace(f"$argCheck[{arg_total}]\n", "")
                return code
        except Exception:
            raise SyntaxError("Not enough arguments in $argCheck!")
    return code


def format_datetime(datetime: d_t, form: str, tz):
    unformated_datetime = datetime.astimezone(tz)
    unformulated_datetime_tuple = (
        unformated_datetime.year,
        unformated_datetime.month,
        unformated_datetime.day,
        unformated_datetime.hour,
        unformated_datetime.minute,
        unformated_datetime.second,
        unformated_datetime.microsecond,
    )
    year, month, day, hour, minute, second, microsecond = unformulated_datetime_tuple

    am_pm = "AM" if hour < 12 else "PM"
    hour = hour if hour < 12 else hour - 12

    form = form.lower().strip()

    if form == "full":
        desired_date_form = f"USA: {month}/{day}/{year} at {hour} :{minute} :{second} :{microsecond} {am_pm}"
    elif form == "year":
        desired_date_form = str(year)
    elif form == "month":
        desired_date_form = str(month)
    elif form == "day":
        desired_date_form = str(day)
    elif form == "hour":
        desired_date_form = str(hour)
    elif form == "minute":
        desired_date_form = str(minute)
    elif form == "second":
        desired_date_form = str(second)
    elif form == "microsecond":
        desired_date_form = str(microsecond)
    elif form == "ampm":
        desired_date_form = am_pm
    else:
        desired_date_form = "ERROR"
    return desired_date_form



from CharmCord.globeHandler import get_globals

async def messageAuthor(ids, context):
    bots = get_globals()[1]

    try:
        args = ids.split(";")
        channel = await bots.fetch_channel(args[0])
        mes = int(args[1])
        message = await channel.fetch_message(mes)
        return message.author
    except:
        raise SyntaxError("Not a message args")



import json
from CharmCord.globeHandler import get_globals


async def getServerVar(args, context):
    all_vars = get_globals()[2]

    ag = args.split(";")
    server = ag[0]
    var = ag[1]
    try:
        with open("variables.json", "r") as variables:
            total = json.load(variables)
            return total[f"{server}_{var}"]
    except KeyError:
        with open("variables.json", "r") as variables:
            total = json.load(variables)
        with open("variables.json", "w") as variables:
            total.update({f"{server}_{var}": all_vars[var]})
            json.dump(total, variables)
            return all_vars[var]



import discord


async def onlyIf(args, context):
    choices = ["==", ">=", "<=", "!=", "<", ">"]
    operator_mapping = {
        "==": lambda x, y: x == y,
        "!=": lambda x, y: x != y,
        ">": lambda x, y: x > y,
        ">=": lambda x, y: x >= y,
        "<": lambda x, y: x < y,
        "<=": lambda x, y: x <= y,
    }
    if ";" in args:
        values = args.split(";")
        for i in choices:
            if i in args:
                if i in ["==", "!="]:
                    vals = values[0].split(i)
                    val1 = vals[0].strip()
                    val2 = vals[1].strip()
                else:
                    vals = values[0].split(i)
                    val1 = int(vals[0])
                    val2 = int(vals[1])
                result = operator_mapping.get(i, lambda x, y: None)(val1, val2)
                if result:
                    return True
                else:
                    if isinstance(context, discord.Interaction):
                        try:
                            await context.response.send_message(values[1])
                        except:
                            await context.followup.send(values[1])
                    else:
                        await context.send(values[1])
                    return False
    else:
        value = args.split(";")
        for i in choices:
            if i in value[0]:
                if i in ["==", "!="]:
                    vals = args.split(i)
                    val1 = vals[0].strip()
                    val2 = vals[1].strip()
                else:
                    vals = args.split(i)
                    val1 = int(vals[0])
                    val2 = int(vals[1])
                result = operator_mapping.get(i, lambda x, y: None)(val1, val2)
                if result:
                    return True
                else:
                    return False



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals



async def guildName(id, context):
    if len(id) < 1:
        raise CharmCordErrors("No parameter provided for '$guildName'")
    bots = get_globals()[1]

    try:
        int(id)
        guild = await bots.fetch_guild(id)
        return guild.name
    except ValueError:
        CharmCordErrors("Invalid ID for '$guildName'")



import discord

from CharmCord.CharmErrorHandling import CharmCordErrors


async def defer(emp, context):
    if isinstance(context, discord.Interaction):
        await context.response.defer()
        return ""
    else:
        CharmCordErrors(f"$defer: Defer cannot be used outside slash commands\nCommand: {context.command.name}")



class CharmErrorHandling:
    @staticmethod
    def Errors(self, error, code):
        """
        ERRORS:

        1: user     args
        2: channel  args
        3: category args

        4: WILDCARD
        """

        if error == 1:
            raise SyntaxError(f"args '{code}' is not a user args")
        if error == 2:
            raise SyntaxError(f"args '{code}' is not a channel args")
        if error == 3:
            raise SyntaxError(f"args '{code}' is not a category args")

        if error == 4:
            raise SyntaxError(code)


class CharmCordErrors(Exception):
    def __init__(self, error: str):
        raise SyntaxError(error)



options = {
    "newChannel": {
        "name": "",
        "id": 0,
        "type": "",
        "category": "",
        "categoryid": 0,
        "guild": "",
        "nsfw": "",
        "delay": 0,
        "created": ""
    },
    "oldChannel": {
        "name": "",
        "id": 0,
        "type": "",
        "category": "",
        "categoryid": 0,
        "guild": "",
        "nsfw": "",
        "delay": 0,
        "created": ""
    },
    "memberJoined": {
        "id": 0,
        'guildid': 0
    },
    "deletedChannel": {
        "name": "",
        "id": 0,
        "type": "",
        "category": "",
        "categoryid": 0,
        "guild": "",
        "nsfw": "",
        "delay": 0,
        "created": ""
    },
    "reactionAdded": {
        "name": "",
        "emoji": "",
        "id": 0,
        "bot_reacted": '',
        "users_reacted": [],
        "count": 0,
        "username": '',
        'msgid': 0,
        'msgauthorid': 0,
        'msgauthorname': '',
        "userid": 0,
    },
    "reactionRemoved": {
        "name": "",
        "emoji": "",
        "id": 0,
        "bot_reacted": '',
        "users_reacted": [],
        "count": 0,
        "username": '',
        'msgid': 0,
        'msgauthorid': 0,
        'msgauthorname': '',
        "userid": 0,
    }
}



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals


async def channelMention(id, context):
    """
    Ex. $channelMention[ChannelID]
    returns a channel mention from the given args
    """
    if len(id) < 1:
        CharmCordErrors("No parameter provided for '$channelMention'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(id.replace("<#", "").replace(">", ""))
        return channel.mention
    except ValueError:
        CharmCordErrors(f"$channelMention: {id} not valid channel id\nCommand: {context.command.name}")



import json


async def getVar(name, context):
    try:
        with open("variables.json", "r") as var:
            total = json.load(var)
            return total[name]
    except KeyError:
        raise SyntaxError(f"{name} variable not set")



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelDelay(id, context):
    """
    Ex. $channelDelay[ChannelID]
    Returns the channel delay for the given channel args
    """
    if len(id) < 1:
        raise CharmCordErrors("No parameter provided for '$channelDelay'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(id.replace("<#", "").replace(">", ""))
        return channel.delay
    except ValueError:
        CharmCordErrors(f"$channelDelay: {id} not valid channel id\nCommand: {context.command.name}")



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelCategoryName(args, context):
    """
    Ex. $channelCategoryName[ChannelID]
    returns the name of the category name of the given args
    """
    if len(args) < 1:
        CharmCordErrors("No parameter provided for '$channelCategoryName'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(args.replace("<#", "").replace(">", ""))
        return channel.category.name
    except ValueError:
        CharmCordErrors(f"$channelCategoryName: {args} not valid channel id\nCommand: {context.command.name}")



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def userID(user: str, context):
    bots = get_globals()[1]

    try:
        user = int(user.replace("<@", "").replace(">", ""))
        new_user = await bots.fetch_user(user)
    except ValueError:
        CharmCordErrors(f"mention ({user}) in $userID not valid")
        return
    return new_user.id



all_Funcs = [
    "$If",
    "$ElIf",
    "$addButton",
    "$buttonSend",
    "$authorID",
    "$botAvatar",
    "$botGuilds",
    "$botID",
    "$botMention",
    "$botName",
    "$channelCategoryID",
    "$channelCategoryName",
    "$channelChangedRoles",
    "$channelCreated",
    "$channelDelay",
    "$channelID",
    "$channelMention",
    "$channelName",
    "$channelNsfw",
    "$channelPosition",
    "$channelType",
    "$channelURL",
    "$charmAI",
    "$console",
    "$count",
    "$defer",
    "$deleteMessage",
    "$deletedChannel",
    "$divide",
    "$editMessage",
    "$get",
    "$getJson",
    "$getVar",
    "$guildID",
    "$guildName",
    "$guildTextChannels",
    "$hasPerm",
    "$let",
    "$memberJoined",
    "$mentions",
    "$message",
    "$messageAuthor",
    "$messageContent",
    "$messageID",
    "$multi",
    "$oldChannel",
    "$onlyIf",
    "$ping",
    "$pyEval",
    "$purge",
    "$reactionAdded",
    "$reactionRemoved",
    "$sendMessage",
    "$sendDM",
    "$sendEmbed",
    "$setServerVar",
    "$setUserVar",
    "$setVar",
    "$slashSend",
    "$sub",
    "$sum",
    "$userMention",
    "$userName",
    "$userID",
    "$wait",
    "$waitMessage",
    "$waitReaction"
]

date_funcs = ["$channelCreated"]

ifse = ["$If", "$ElIf"]

no_arg_Funcs = ["$botName",
                "$botID",
                "$authorID",
                "$channelID",
                "$guildID",
                "$ping",
                "$botGuilds",
                "$botAvatar",
                "$botMention",
                "$messageID",
                "$message",
                "$defer"]

newline_char = "*n*"



import json
import discord
from discord.ext import commands
from CharmCord.functions.Events._options_ import options
from CharmCord.functions.Messages._btnOpts_ import button
from CharmCord.tools import FunctionHandler, findBracketPairs, noArguments
from .CommandHandler import load_commands
from .Commands import Commands
from .SlashCommands import SlashCommands
from ..globeHandler import update_globals, get_globals

global bots


class CharmCord:

    def __init__(
            self,
            prefix,
            case_insensitive,
            intents,
            activity,
            load_command_dir,
    ):
        """
        This function initializes a bot with specified prefix, case sensitivity, intents, activity, and loads command files.
        It determines intents based on user input and creates bot instances with the specified parameters.
        Additionally, it loads command files and sets up a JSON file for storing variables if it doesn't exist.
        :param prefix:
        :param case_insensitive:
        :param intents:
        :param activity:
        :param load_command_dir:
        """
        # Global variables
        global bots

        # Initialize Start class
        self.prefix = prefix
        self.case_insensitive = case_insensitive
        self.intented = intents
        self._help_command = None
        self._clients = ""
        self.intent = ""
        self._activity = activity
        self.all_variables = {}
        self.bot = None

        # Determine intents
        if "all" in self.intented:
            self.intent = discord.Intents.all()
        elif "default" in self.intented:
            self.intent = discord.Intents.default()
        else:
            self.intent = discord.Intents.default()

        # Enable specific intents
        if "message" in self.intented:
            self.intent.message_content = True
        if "members" in self.intented:
            self.intent.members = True
        if "presences" in self.intented:
            self.intent.presences = True

        # Create bot instances

        if self._activity is None:
            self._clients = commands.Bot(
                command_prefix=self.prefix,
                case_insensitive=self.case_insensitive,
                intents=self.intent,
                help_command=self._help_command,
            )
        else:
            self._clients = commands.Bot(
                command_prefix=self.prefix,
                case_insensitive=self.case_insensitive,
                intents=self.intent,
                activity=self._activity,
                help_command=self._help_command,
            )
        bots, self.bot = self._clients, self._clients
        update_globals("bots", self.bot)

        try:
            load_commands(load_command_dir)
        except FileNotFoundError:
            pass

        try:
            with open("variables.json", "r") as var:
                pass
        except FileNotFoundError:
            with open("variables.json", "w") as var:
                go = {"STRD": True}
                json.dump(go, var)

    @staticmethod
    def run(token: str):
        bots.run(token)

    def variables(self, variables: dict):
        for key, value in variables.items():
            self.all_variables[key] = value
        all_vars = self.all_variables
        update_globals("all", all_vars)

    @staticmethod
    def button_code(name, code):
        button[name] = code
        return

    @staticmethod
    def slash_command(name: str, code: str, args: list[dict] = None, description: str = "") -> None:
        sl = SlashCommands().slash_command
        sl(name=name, code=code, args=args, description=description.lower(), bot=bots)

    @staticmethod
    def command(name: str, code: str, aliases: list = None):
        co = Commands().command
        if aliases is None:
            co(name=name, code=code, bot=bots)
        else:
            co(name=name, code=code, aliases=aliases, bot=bots)

    # EVENTS BELOW

    def on_reaction_add(self, code=None):
        @self.bot.event
        async def on_reaction_add(reaction: discord.Reaction, user: discord.User):
            try:
                options["reactionAdded"]["name"] = reaction.emoji.name
                options["reactionAdded"]["id"] = reaction.emoji.id
            except AttributeError:
                options["reactionAdded"]["name"] = reaction.emoji  # May Remove
                options["reactionAdded"]["id"] = 0
            options["reactionAdded"]["bot_reacted"] = reaction.me
            options["reactionAdded"]["users_reacted"] = [user async for user in reaction.users()]
            options["reactionAdded"]["count"] = reaction.count
            options["reactionAdded"]["username"] = user.name
            options["reactionAdded"]["userid"] = user.id
            options["reactionAdded"]["msgid"] = reaction.message.id
            options["reactionAdded"]["msgauthorid"] = reaction.message.author.id
            options["reactionAdded"]["msgauthorname"] = reaction.message.author.name

            if code is not None:
                final_code = await noArguments(code, TotalFuncs, None)
                await findBracketPairs(final_code, TotalFuncs, None)
            return

    def on_reaction_remove(self, code=None):
        @self.bot.event
        async def on_reaction_remove(reaction: discord.Reaction, user: discord.User):
            try:
                options["reactionRemoved"]["name"] = reaction.emoji.name
                options["reactionRemoved"]["id"] = reaction.emoji.id
            except AttributeError:
                options["reactionRemoved"]["name"] = reaction.emoji  # May Remove
                options["reactionRemoved"]["id"] = 0
            options["reactionRemoved"]["bot_reacted"] = reaction.me
            options["reactionRemoved"]["users_reacted"] = [user async for user in reaction.users()]
            options["reactionRemoved"]["count"] = reaction.count
            options["reactionRemoved"]["username"] = user.name
            options["reactionRemoved"]["userid"] = user.id
            options["reactionRemoved"]["msgid"] = reaction.message.id
            options["reactionRemoved"]["msgauthorid"] = reaction.message.author.id
            options["reactionRemoved"]["msgauthorname"] = reaction.message.author.name

            if code is not None:
                final_code = await noArguments(code, TotalFuncs, None)
                await findBracketPairs(final_code, TotalFuncs, None)
            return

    def on_message(self, code=None):
        @self.bot.event
        async def on_message(msg: discord.Message):
            pass  # For now

    def on_member_join(self, code=None):
        @self.bot.event
        async def on_member_join(member: discord.Member):
            options["memberJoined"]["id"] = member.id
            options["memberJoined"]["guildid"] = member.guild.id

            if code is not None:
                final_code = await noArguments(code, TotalFuncs, None)
                await findBracketPairs(final_code, TotalFuncs, None)
            return

    def on_channel_updated(self, code=None):
        @self.bot.event
        async def on_guild_channel_update(before, after):
            for i in options["oldChannel"].keys():
                options["oldChannel"][i] = before.i
            for i in options["newChannel"].keys():
                options["newChannel"][i] = after.i
            if code is not None:
                final_code = await noArguments(code, TotalFuncs, None)
                await findBracketPairs(final_code, TotalFuncs, None)

    def on_channel_deleted(self, code=None):
        @self.bot.event
        async def on_guild_channel_delete(channel):
            options["deletedChannel"]["name"] = channel.name
            options["deletedChannel"]["id"] = channel.id
            options["deletedChannel"]["type"] = channel.type
            options["deletedChannel"]["category"] = channel.category
            options["deletedChannel"]["categoryid"] = channel.category_id
            options["deletedChannel"]["guild"] = channel.guild
            options["deletedChannel"]["nsfw"] = channel.nsfw
            options["deletedChannel"]["delay"] = channel.slowmode_delay
            options["deletedChannel"]["created"] = channel.created_at
            # more options coming
            if code is not None:
                final_code = await noArguments(code, TotalFuncs, None)
                await findBracketPairs(final_code, TotalFuncs, None)

    def on_ready(self, code):
        @self.bot.event
        async def on_ready():
            from CharmCord.CharmErrorHandling import CharmCordErrors
            final_code = await noArguments(code, TotalFuncs, None)
            await findBracketPairs(final_code, TotalFuncs, None)
            try:
                await self.bot.tree.sync()
            except Exception as e:
                print(e)
                CharmCordErrors("All slash commands need a description")


def CharmClient(
        prefix: str,
        case_insensitive: bool = False,
        intents: tuple = ("default",),
        activity=None,
        load_command_dir="commands",
):
    """
    CharmCord Discord Client
    """
    # Global variables
    global bots
    global TotalFuncs

    # Initialize FunctionHandler and register functions
    functions = FunctionHandler()
    TotalFuncs = functions
    update_globals("total", functions)
    functions.register_functions()

    # Create Start instance and return working bot
    _final = CharmCord(
        prefix,
        case_insensitive,
        intents,
        activity,
        load_command_dir
    )
    return _final



import discord
from CharmCord.CharmErrorHandling import CharmCordErrors

count = 0


async def purge(args, context):
    global count

    def check(msg):
        global count
        from CharmCord.utils.CharmCord import bots
        if isinstance(context, discord.Interaction):
            count += 1
            if msg.author.id != bots.user.id or count > 1:
                return True
        else:
            if context.message.id != msg.id:
                return True

    from CharmCord.utils.CharmCord import bots
    vals = args.split(";")
    channel = vals[0]
    amount = vals[1]
    try:
        checking = vals[2].lower()
    except IndexError:
        checking = "false"
    try:
        channel = await bots.fetch_channel(channel)
    except:
        CharmCordErrors(f"{channel} not a channel ID or Name")

    if checking == "true":
        try:
            if isinstance(context, discord.Interaction):
                await context.response.defer()
                await channel.purge(limit=int(amount) + 1, check=check)
                count = 0
                return
            await channel.purge(limit=int(amount), check=check)
            count = 0
            return
        except ValueError:
            CharmCordErrors(f"{amount} is not a digit! Needs to be digit")
    elif checking == "false":
        try:
            await channel.purge(limit=int(amount))
        except ValueError:
            CharmCordErrors(f"{amount} is not a digit! Needs to be digit")



import json


async def setVar(args, context):
    pass

    ag = args.split(";")
    try:
        with open("variables.json", "r") as var:
            total = json.load(var)
            try:
                total.update({ag[0]: int(ag[1])})
            except:
                total.update({ag[0]: ag[1]})
        with open("variables.json", "w") as var:
            json.dump(total, var)
    except KeyError:
        raise SyntaxError(f"{ag[0]} variable not set")



import aiohttp


async def charmAI(args, context):
    api_token = '...'

    # Set up the API base URL
    base_url = 'https://api.gitbook.com/v1'
    space_id = "zxDSJg6koIjslgO2E0IN"

    # Define the request body as a dictionary
    request_body = {
        "query": args,
        # Add other search parameters as needed
    }

    # Use aiohttp to make an asynchronous POST request
    async with aiohttp.ClientSession() as session:
        async with session.post(f'{base_url}/spaces/{space_id}/search/ask',
                                json=request_body,
                                headers={'Authorization': f'Bearer {api_token}'}) as response:
            if response.status == 200:
                books_data = await response.json()
                try:
                    answer_text = books_data["answer"]["text"]
                    return answer_text
                except KeyError:
                    return "No Answer.. | Try again"
            else:
                return None



from CharmCord.CharmErrorHandling import CharmCordErrors
from CharmCord.globeHandler import get_globals

async def channelURL(id, context):
    """
    Ex. $channelURL
    returns the channel url to the given args
    """
    if len(id) < 1:
        CharmCordErrors("No parameter provided for '$channelURL'")
    bots = get_globals()[1]

    try:
        channel = await bots.fetch_channel(id.replace("<#", "").replace(">", ""))
        return channel.jump_url
    except ValueError:
        CharmCordErrors(f"$channelURL: {id} not valid channel id\nCommand: {context.command.name}")



import discord

from CharmCord.all_functions import newline_char


async def slashSend(args: str, context: discord.Interaction):
    try:
        message = args
        message = message.replace(newline_char, "\n")
        try:
            await context.response.send_message(message)
        except Exception:
            await context.followup.send(message)
    except Exception as e:
        raise SyntaxError("Can't send empty message!")
    return
